/* Combigreater
   the greater parser combinator library
   by Danny Cook */

'use strict'

/* helpers */

function success(rest, value) {
    return { succeeded  : true
           , value      : value
           , rest       : rest
           }
}

function failure(rest, expected) {
    return { succeeded  : false
           , expected   : expected === undefined ? null : expected
           , rest       : rest
           }
}

var liberate    = require('liberate')
  , sliceArray  = liberate(Array.prototype.slice)
  , sliceString = liberate(String.prototype.slice)
  , mapArray    = liberate(Array.prototype.map)
  , concat      = liberate(Array.prototype.concat)
  , reduce      = liberate(Array.prototype.reduce)
  , every       = liberate(Array.prototype.every)
  , some        = liberate(Array.prototype.some)

function repeat(count, value) {
    return Array(count).join(' ').split(' ').map(function () { return value })
}

/* simple parsers */

var nothing = expect('nothing', function (input) {
    return success(input, null)
})
  , fail    = expect('fail', function (input) {
    return failure(input)
})
  , end     = expect('end of input', function (input) {
    return input.length === 0 ? success(input, null)
           /* otherwise */    : failure(input)
})

function symbol(s) {
    return expect('the symbol ' + s, matching(function (s_) {
        return s_ === s
    }))
}

function symbols(ss) {
    return map(function () {
        return ss
    }, expect('the symbols "' + ss + '"', all(mapArray(ss, symbol))))
}

function oneOf(ss) {
    return expect('one of "' + ss + '"', any(mapArray(ss, symbol)))
}

function matching(predicate) {
    return expect('a symbol matching the predicate ' + predicate.toString(),
    function (input) {
        var s = input[0]
        if (predicate(s)) {
            return success((typeof input === 'string' ? sliceString : sliceArray)(input, 1), s)
        } else {
            return failure(input)
        }
    })
}

/* parser combinators */

function map(transform, parser) {
    return expect(parser.expected, function (input) {
        var result = parser(input)
        if (result.succeeded) result.value = transform(result.value)
        return result
    })
}

function expect(expected, parser) {
    var newParser = function (input) {
        var result = parser(input)
        if (!result.succeeded) result.expected = expected
        return result
    }

    newParser.expected = expected
    return newParser
}

function lazy(createParser, args) {
    return function (input) {
        return createParser.apply(this, args)(input)
    }
}

function optional(parser) {
    return expect('optionally ' + parser.expected, function (input) {
        var result = parser(input)
        return result.succeeded ? success(result.rest, result.value)
               /* otherwise */  : success(input      , null        )
    })
}

function many(parser) {
    return expect(parser.expected + ' zero or more times', function (input) {
        var values = []
          , nextInput = input
          , result

        while ((result = parser(nextInput)).succeeded) {
            values.push(result.value)
            nextInput = result.rest
        }

        return success(result.rest, values)
    })
}

function many1(parser) {
    return expect(parser.expected + ' one or more times', map(function (pair) {
        return concat([pair[0]], pair[1])
    }, all([parser, many(parser)])))
}

function not(parser) {
    return expect('not ' + parser.expected, function (input) {
        return parser(input).succeeded ? failure(input)
                                       : success(input, null)
    })
}

function all(parsers) {
    return function (input) {
        var values = []
          , nextInput = input
          , result
          , everySucceeded = every(parsers, function (parser) {
            result = parser(nextInput)

            if (result.succeeded) {
                values.push(result.value)
                nextInput = result.rest
                return true
            } else {
                return false
            }
        })
        
        return everySucceeded  ? success(nextInput, values         )
               /* otherwise */ : failure(input    , result.expected)
    }
}

function any(parsers) {
    function getExpected(parser) {
        return parser.expected
    }

    return expect(mapArray(parsers, getExpected).join(' or '),
    function (input) {
        var result
          , someSucceeded = some(parsers, function (parser) {
            return (result = parser(input)).succeeded
        })

        return someSucceeded   ? result
               /* otherwise */ : failure(input)
    })
}

function times(count, parser) {
    return expect(parser.expected + ' ' + count.toString() + ' times',
                  all(repeat(count, parser)))
}

/* exports */

module.exports = { nothing  : nothing
                 , fail     : fail
                 , end      : end
                 , symbol   : symbol
                 , symbols  : symbols
                 , oneOf    : oneOf
                 , matching : matching

                 , map      : map
                 , expect   : expect
                 , lazy     : lazy
                 , optional : optional
                 , many     : many
                 , many1    : many1
                 , not      : not
                 , all      : all
                 , any      : any
                 , times    : times
                 }
